package io.tesla.webserver;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import javax.inject.Named;
import javax.inject.Singleton;
import javax.net.ssl.SSLContext;
import javax.servlet.Servlet;

import org.eclipse.jetty.security.ConstraintMapping;
import org.eclipse.jetty.security.ConstraintSecurityHandler;
import org.eclipse.jetty.security.HashLoginService;
import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.ConnectHandler;
import org.eclipse.jetty.server.handler.ContextHandler;
import org.eclipse.jetty.server.handler.ContextHandlerCollection;
import org.eclipse.jetty.server.handler.ResourceHandler;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.server.ssl.SslSelectChannelConnector;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.eclipse.jetty.servlets.ProxyServlet;
import org.eclipse.jetty.util.resource.Resource;
import org.eclipse.jetty.util.security.Constraint;
import org.eclipse.jetty.util.security.Password;
import org.eclipse.jetty.util.ssl.SslContextFactory;
import org.eclipse.jetty.webapp.WebAppClassLoader;
import org.eclipse.jetty.webapp.WebAppContext;

// eclipse
// command line like workspace
// deployed as war
// deployed as incremental

@Singleton
@Named("jetty8")
public class Jetty8WebServer implements WebServer {

  private Server server;
  private Runtime runtime;
  private ContextHandlerCollection contexts;
  private int port;
  private String host;
  private String protocol;
  private boolean enableSsl;
  private ServletContextHandler servletContextHandler;
  private SSLContext sslContext;

  private ConstraintSecurityHandler securityHandler = new ConstraintSecurityHandler();
  private HashLoginService loginService;
  private String authType;

  //
  // SSL
  //
  private File keyStore;
  private String keyStorePassword;

  public Jetty8WebServer() {
    this(0);
  }

  public Jetty8WebServer(int port) {
    this.port = port;
    this.runtime = new Runtime();

    server = new Server();
    contexts = new ContextHandlerCollection();
    servletContextHandler = new ServletContextHandler(contexts, "/", ServletContextHandler.SESSIONS);
  }

  public File getWorkDirectory() {
    return runtime.getWorkDirectory();
  }

  public void enableSsl(SSLContext sslContext) {
    this.enableSsl = true;
    this.sslContext = sslContext;
  }

  public void enableSsl(File keyStore, String keyStorePassword) {
    this.enableSsl = true;
    this.keyStore = keyStore;
    this.keyStorePassword = keyStorePassword;
  }

  public void run() throws Exception {
    start();
    join();
  }

  public void start() throws Exception {

    Connector connector;

    if (enableSsl) {
      SslContextFactory sslContextFactory = new SslContextFactory();
      if (sslContext == null) {

      } else {
        System.out.println("Using provided SSLContext");
        System.out.println(sslContext);
        sslContextFactory.setSslContext(sslContext);
      }
      connector = new SslSelectChannelConnector();
      ((SslSelectChannelConnector) connector).setSslContext(sslContext);
    } else {
      connector = new SelectChannelConnector();
    }

    connector.setPort(port);
    server.addConnector(connector);
    connector.open();

    server.setHandler(contexts);
    server.start();

    if (enableSsl) {
      protocol = "https";
    } else {
      protocol = "http";
    }

    port = connector.getLocalPort();
    host = connector.getHost();
  }

  public void join() throws Exception {
    server.join();
  }

  public void stop() throws Exception {
    server.stop();
  }

  public void addServlet(Servlet servlet, String pathSpecification) {
    ServletContextHandler context = new ServletContextHandler(ServletContextHandler.SESSIONS);
    context.setContextPath("/");
    context.addServlet(new ServletHolder(servlet), pathSpecification);
    contexts.addHandler(context);
  }

  public void addServlet(Servlet servlet, String contextPath, String pathSpecification) {
    ServletContextHandler context = new ServletContextHandler(ServletContextHandler.SESSIONS);
    context.setContextPath(contextPath);
    context.addServlet(new ServletHolder(servlet), pathSpecification);
    contexts.addHandler(context);
  }

  public void addResourceContext(String context, String directory) throws MalformedURLException, IOException {
    File resourceDirectory = new File(runtime.getCurrentWorkingDirectory(), directory);
    addResourceContext(context, resourceDirectory);
  }

  public void addResourceContext(String context, File directory) throws MalformedURLException, IOException {
    ContextHandler contextHandler = new ContextHandler();
    ResourceHandler contentResourceHandler = new ResourceHandler();
    contextHandler.setBaseResource(Resource.newResource(directory));
    contextHandler.setHandler(contentResourceHandler);
    contextHandler.setContextPath(context);
    addContextHandler(contextHandler);
  }

  public void addWebAppContext(WebAppContext contextHandler) {
    contexts.addHandler(contextHandler);
  }

  protected void addContextHandler(ContextHandler contextHandler) {
    contexts.addHandler(contextHandler);
  }

  public void addWebAppContext(File webAppDirectory, String context) throws MalformedURLException, IOException {
    addContextHandler(new WebAppContext(webAppDirectory.getAbsolutePath(), context));
  }

  public void addWebAppContext(String context) throws MalformedURLException, IOException {

    if (isInSitu(runtime.getCurrentWorkingDirectory())) {
      System.out.println("IN SITU");
      addContextHandler(inSituWebApp(context, runtime.getCurrentWorkingDirectory()));
    } else if (isExplodedWebApp(runtime.getCurrentWorkingDirectory())) {
      System.out.println("EXPLODED");
      addContextHandler(explodedWebApp(runtime.getCurrentWorkingDirectory()));
    } else if (isWarWebApp(runtime.getCurrentWorkingDirectory())) {
      System.out.println("WAR");
      //addContextHandler(warWebApp(currentWorkingDirectory));
    }
  }

  //
  // In Situ
  //
  private boolean isInSitu(File currentWorkingDirectory) {
    if (new File(currentWorkingDirectory, "pom.xml").exists()) {
      //
      // A process that starts up from the directory where the POM is like an IDE launch configuration,
      // or a uber JAR that starts up a webserver.
      //
      return true;
    } else if (new File(currentWorkingDirectory, "../../../pom.xml").exists()) {
      //
      // An OSX process being started from an application launcher where the startup script
      // is three directories down:
      //
      // Idiom.app
      // └── Contents
      //     ├── Info.plist
      //     ├── MacOS
      //     │   └── Run
      //     └── Resources
      //         ├── AppSettings.plist
      //         ├── MainMenu.nib
      //         │   ├── designable.nib
      //         │   └── keyedobjects.nib
      //         ├── appIcon.icns
      //         └── script
      //
      // Where the script file may contain something like the following:
      //
      // #!/bin/sh
      //
      // java -jar ../../../idiom-designer-0.0.1-SNAPSHOT.jar
      //
      // Where we are walking back up three directories so the working directory is not the base of
      // the project so we need to adjust for that.
      //      
      return true;
    }
    return false;
  }

  private WebAppContext inSituWebApp(String context, File currentWorkingDirectory) throws IOException {

    WebAppContext webAppContext = new WebAppContext();
    webAppContext.setContextPath(context);
    File webAppDirectory = new File(currentWorkingDirectory, "src/main/webapp");

    if (!webAppDirectory.exists()) {
      throw new RuntimeException("Expecting to find a standard webapp project but there is no src/main/webapp directory.");
    }

    webAppContext.setBaseResource(Resource.newResource(webAppDirectory));
    webAppContext.setParentLoaderPriority(true);

    return webAppContext;
  }

  //
  // Exploded 
  //  
  private boolean isExplodedWebApp(File currentWorkingDirectory) {
    File webAppDirectory = new File(currentWorkingDirectory, "../webapps/idiom");
    return new File(webAppDirectory, "classes").exists();
  }

  private WebAppContext explodedWebApp(File currentWorkingDirectory) throws IOException {

    File webAppDirectory = new File(currentWorkingDirectory, "../webapps/idiom");

    //
    // This is the deployment structure which separates the classes that belong to the app itself, the
    // libraries that the app uses, and the resources of the webapp.
    //
    // webapps
    // └── context
    //   ├── classes
    //   ├── jars
    //   └── webapp
    //
    File webAppClasses = new File(webAppDirectory, "classes");
    File webAppJars = new File(webAppDirectory, "jars");
    File webAppResources = new File(webAppDirectory, "webapp");
    //
    // The context for the webapp is the directory name the webapp is contained in.
    //
    WebAppContext webAppContext = new WebAppContext();
    webAppContext.setContextPath("/" + webAppDirectory.getName());
    //
    // Setup the classloader for the webapp
    //
    WebAppClassLoader webAppClassLoader = new WebAppClassLoader(webAppContext);
    webAppClassLoader.addJars(Resource.newResource(webAppJars));
    webAppClassLoader.addClassPath(Resource.newResource(webAppClasses));
    webAppContext.setClassLoader(webAppClassLoader);
    webAppContext.setParentLoaderPriority(true);
    //
    //
    //
    webAppContext.setBaseResource(Resource.newResource(webAppResources));

    return webAppContext;
  }

  //
  // War
  //
  private boolean isWarWebApp(File currentWorkingDirectory) throws IOException {

    File webAppsDirectory = new File(currentWorkingDirectory, "webapps");
    System.out.println(webAppsDirectory);
    String[] files = webAppsDirectory.list();
    for (String file : files) {
      //
      // Look for webapps to deploy
      //
      File webappDirectory = new File(webAppsDirectory, file);
      if (webappDirectory.isDirectory()) {
        File[] f = webappDirectory.listFiles();
        if (f != null && f.length == 1) {
          if (f[0].getName().endsWith(".war")) {
            System.out.println(f[0]);
            addContextHandler(warWebApp("/" + webappDirectory.getName(), f[0]));
            return true;
          }
        }
      }
    }
    return false;
  }

  private WebAppContext warWebApp(String context, File war) throws IOException {
    WebAppContext webAppContext = new WebAppContext();
    webAppContext.setContextPath(context);
    webAppContext.setWar(war.getAbsolutePath());
    webAppContext.setParentLoaderPriority(true);
    return webAppContext;
  }

  public int getPort() {
    return port;
  }

  public String getHost() {
    return host;
  }

  public String getProtocol() {
    return protocol;
  }

  public URL getUrl() throws MalformedURLException {
    if (host == null) {
      host = "localhost";
    }
    return new URL(protocol, host, port, "");
  }

  //
  // Security
  //
  public void addProxySupport() {

  }

  public void addAuthentication(String pathSpec, String authName) {
    if (server == null) {
      try {
      } catch (Exception e) {
        throw new IllegalStateException(e);
      }
    }
    initAuthentication(pathSpec, authName);
  }

  private void initAuthentication(String pathSpec, String authName) {
    authType = authName;
    Constraint constraint = new Constraint();
    if (authName == null) {
      authName = Constraint.__BASIC_AUTH;
    }
    constraint.setName(authName);

    constraint.setRoles(new String[] {
      "users"
    });
    constraint.setAuthenticate(true);

    ConstraintMapping cm = new ConstraintMapping();
    cm.setConstraint(constraint);
    cm.setPathSpec(pathSpec);

    securityHandler.setRealmName("Test Server");
    securityHandler.setAuthMethod(authName);
    securityHandler.setStrict(true);

    securityHandler.setConstraintMappings(new ConstraintMapping[] {
      cm
    });
    loginService = new HashLoginService("Test Server");
    securityHandler.setLoginService(loginService);
    servletContextHandler.setSecurityHandler(securityHandler);
  }

  /**
   * Add the given user to the LoginService. If the password object is a {@link CertificateHolder},
   * {@link #addCertificate(String, CertificateHolder)} is called. For any other class, the String representation of
   * the object is used as a password.
   * 
   * @param user the username, may not be {@code null}.
   * @param password The password to use, may not be {@code null}.
   */
  public void addUser(String user, Object password) {
    if (authType == null) {
      throw new IllegalStateException("no authentication method set.");
    }
    loginService.putUser(user, new Password(password.toString()), new String[] {
      "users"
    });
  }

  @Override
  public void enableProxy() {
    
    System.out.println("Enabling proxies!");
    addServlet(new ProxyServlet(), "/");

    ConnectHandler proxy = new ConnectHandler();
    proxy.setWhite(new String[] { "mail.google.com" });
    proxy.addWhite("www.google.com");
    contexts.addHandler(proxy);
  }
}
